use std::sync::Arc;

use axum::response::IntoResponse;
use config::Config;
use http::{HeaderMap, HeaderValue, StatusCode};
use sqlx::{Pool, Postgres};
use thiserror::Error;

use crate::model::{Chat, Message};

type Result<T> = std::result::Result<T, AppStateError>;
const DB_URL: &str = "postgres://chatdings@localhost/chatdings";

#[derive(Debug, Clone)]
pub struct AppState(Arc<AppStateInner>);

#[derive(Debug)]
pub struct AppStateInner {
    pool: Pool<Postgres>,
    config: Config,
}

mod config {
    use std::{env, fs};

    #[derive(Debug)]
    pub(super) struct Config {
        db_url: String,
        admin_token: String,
    }

    impl Config {
        pub fn read_from_env() -> Self {
            let db_url = env::var("CHATDINGS_DB_URL")
                .expect("environment variable CHATDINGS_DB_URL has to exist");
            let admin_token_path = env::var("CHATDINGS_ADMIN_TOKEN_PATH")
                .expect("environment variable CHATDINGS_ADMIN_TOKEN has to exist");
            let admin_token = fs::read_to_string(&admin_token_path)
                .expect(&format!("failed to read '{admin_token_path:?}'"))
                .trim()
                .to_string();

            Self {
                db_url,
                admin_token,
            }
        }

        pub fn db_url(&self) -> &str {
            &self.db_url
        }

        pub fn admin_tok(&self) -> &str {
            &self.admin_token
        }
    }
}

impl AppState {
    pub async fn init() -> Result<Self> {
        Ok(Self(Arc::new(AppStateInner::init().await?)))
    }
    pub async fn fetch_chat_by_url_path(&self, url_path: &str) -> Result<Chat> {
        self.0.fetch_chat_by_url_path(url_path).await
    }
    pub async fn fetch_messages(&self, chat: &Chat) -> Result<Vec<Message>> {
        self.0.fetch_messages(chat).await
    }
    pub async fn send_message(&self, chat: &Chat, content: String, from_admin: bool) -> Result<()> {
        self.0.send_message(chat, content, from_admin).await
    }
    pub fn check_admin_tok(&self, headers: &HeaderMap) -> bool {
        self.0.check_admin_tok(headers)
    }
    pub fn pool(&self) -> &Pool<Postgres> {
        self.0.pool()
    }
}

impl AppStateInner {
    async fn init() -> Result<Self> {
        let config = Config::read_from_env();
        let pool = Pool::<Postgres>::connect(config.db_url()).await?;

        sqlx::migrate!()
            .run(&pool)
            .await
            .expect("migration should not fail");

        Ok(Self { pool, config })
    }

    async fn fetch_chat_by_url_path(&self, url_path: &str) -> Result<Chat> {
        Ok(
            sqlx::query_as!(Chat, r#"select * from chats where url_path = $1"#, url_path)
                .fetch_one(&self.pool)
                .await?,
        )
    }

    async fn fetch_messages(&self, chat: &Chat) -> Result<Vec<Message>> {
        Ok(sqlx::query_as!(
            Message,
            r#"select * from messages where chat_id = $1"#,
            chat.id
        )
        .fetch_all(&self.pool)
        .await?)
    }

    async fn send_message(&self, chat: &Chat, content: String, from_admin: bool) -> Result<()> {
        todo!()
    }

    fn check_admin_tok(&self, headers: &HeaderMap) -> bool {
        headers.get("x-admin-tok") == Some(&HeaderValue::from_str(self.config.admin_tok()).unwrap())
    }

    fn pool(&self) -> &Pool<Postgres> {
        &self.pool
    }
}

#[derive(Error, Debug)]
pub enum AppStateError {
    #[error("database error")]
    Sqlx(#[from] sqlx::Error),
}

impl IntoResponse for AppStateError {
    fn into_response(self) -> axum::response::Response {
        todo!()
    }
}

impl From<AppStateError> for StatusCode {
    fn from(value: AppStateError) -> Self {
        match value {
            AppStateError::Sqlx(sqlx::Error::RowNotFound) => Self::NOT_FOUND,
            AppStateError::Sqlx(_) => Self::INTERNAL_SERVER_ERROR,
        }
    }
}
